import json
import os
from pathlib import Path

import typer
from typing_extensions import Annotated
from typing import Optional

from aperturedb.cli.console import console
import aperturedb.cli.keys as keys
from aperturedb.Configuration import Configuration
from aperturedb.CommonLibrary import _create_configuration_from_json, __create_connector


class ObjEncoder(json.JSONEncoder):
    """
    A bit of boiler plate to allow us to serialize our Configuration object.
    """

    def default(self, o):
        return o.__dict__


APP_FULL_NAME = "CLI for aperturedb"
APP_NAME = "aperturedb"
APP_NAME_CLI = "adb"

app = typer.Typer()


def _config_file_path(as_global: bool):
    config_path: Path = Path(os.path.join(
        os.getcwd(),
        f".{APP_NAME}",
        f"{APP_NAME_CLI}.json"))
    if as_global:
        app_dir = typer.get_app_dir(APP_NAME)
        config_path: Path = Path(app_dir) / f"{APP_NAME_CLI}.json"
    return config_path


def has_environment_configuration():
    for known_variable in ["APERTUREDB_KEY", "APERTUREDB_JSON"]:
        if (data := os.environ.get(known_variable)) is not None and data != "":
            return True
    return False


def check_configured(as_global: bool, show_error: bool = False):
    config_path = _config_file_path(as_global)
    file_exists = config_path.exists()
    if show_error:
        recommend = "Please run adb config create <name>"
        if not file_exists:
            console.log(f"{APP_NAME_CLI} is not configured yet. {recommend}")
    return file_exists


def get_configurations(file: str):
    configs = {}
    active = None
    with open(file) as config_file:
        configurations = json.loads(config_file.read())
        for c in filter(lambda key: key != "active", configurations):
            config = configurations[c]
            configs[c] = Configuration(
                name=c,
                host=config["host"],
                port=config["port"],
                username=config["username"],
                password=config["password"],
                use_rest=config["use_rest"],
                use_ssl=config["use_ssl"])
            if "user_keys" in config:
                configs[c].set_user_keys(config["user_keys"])
    active = configurations["active"]
    return configs, active


@app.command()
def ls(log_to_console: bool = True):
    """
    List the configurations with their details.
    """
    all_configs = {}
    for as_global in [True, False]:
        config_path = _config_file_path(as_global)
        context = "global" if as_global else "local"
        try:
            configs, active = get_configurations(config_path.as_posix())
            all_configs[context] = configs
            all_configs["active"] = active
        except FileNotFoundError:
            check_configured(as_global)
        except json.JSONDecodeError:
            check_configured(as_global)
            console.log(f"Failed to decode json '{config_path.as_posix()}'")

    for env_key, plain_json in [["APERTUREDB_JSON", True], ["APERTUREDB_KEY", False]]:
        if (data := os.environ.get(env_key)) is not None and data != "":
            if plain_json:
                config = _create_configuration_from_json(data)
            else:
                config = Configuration.reinflate(data)
            if not "environment" in all_configs:
                all_configs["environment"] = {}
            all_configs["environment"][env_key] = config
            all_configs["active"] = f"env:{env_key}"

    if "global" in all_configs or "local" in all_configs:
        if "global" in all_configs and len(all_configs["global"]) == 0 \
                and "local" in all_configs and len(all_configs["local"]) == 0:
            console.log(
                f"No configurations found. Please run adb config create <name>")
            return
        else:
            if log_to_console:
                console.log(f"Available configurations:")
                console.log(all_configs)

    else:
        console.log(all_configs)
        # Tried to parse global config as well as local, but failed.
        # Show user the error and bail.
        check_configured(as_global=True, show_error=True)

    return all_configs


@app.command()
def create(
        name: Annotated[Optional[str], typer.Argument(
            help="Name of this configuration for easy reference")] = None,
        active: Annotated[bool, typer.Option(help="Set as active")] = False,
        as_global: Annotated[bool, typer.Option(
            help="Project level vs global level")] = True,
        host: Annotated[str, typer.Option(help="Host name")] = "localhost",
        port: Annotated[int, typer.Option(help="Port number")] = 55555,
        username: Annotated[str, typer.Option(help="Username")] = "admin",
        password: Annotated[str, typer.Option(help="Password")] = "admin",
        use_rest: Annotated[bool, typer.Option(help="Use REST")] = False,
        use_ssl: Annotated[bool, typer.Option(help="Use SSL")] = True,
        interactive: Annotated[bool, typer.Option(
            help="Interactive mode")] = True,
        overwrite: Annotated[bool, typer.Option(
            help="overwrite existing configuration")] = False,
        from_json: Annotated[bool, typer.Option(
            help="create config from a JSON string")] = False,
        from_key: Annotated[bool, typer.Option(help="create config from an encoded string")] = False):
    """
    Create a new configuration for the client.

    If --from-json or --from-key is used, then the options --host, --port, --username, --password, --use-rest, and --use-ssl will be ignored.
    The user will be prompted to enter the JSON string.
    This will be treated as a password entry.

    See https://docs.aperturedata.dev/Setup/client/configuration for more information on JSON configurations.
    """

    def check_for_overwrite(name):
        if name in configs and not overwrite:
            console.log(
                f"Configuration named '{name}' already exists. Use --overwrite to overwrite.",
                style="bold yellow")
            raise typer.Exit(code=2)

    db_host = host
    db_port = port
    db_username = username
    db_password = password
    db_use_rest = use_rest
    db_use_ssl = use_ssl

    config_path = _config_file_path(as_global)
    configs = {}
    try:
        configs, ac = get_configurations(config_path.as_posix())
    except FileNotFoundError as e:
        config_path: Path = _config_file_path(as_global)
        parent_dir = os.path.dirname(config_path)
        if not os.path.exists(parent_dir):
            os.makedirs(parent_dir)
        active = True
    except json.JSONDecodeError:
        active = True

    if name is not None:
        check_for_overwrite(name)

    if from_json:
        assert interactive, "Interactive mode must be enabled for --from-json"
        json_str = typer.prompt("Enter JSON string", hide_input=True)
        gen_config = _create_configuration_from_json(
            json_str, name=name, name_required=True)
        check_for_overwrite(gen_config.name)
        name = gen_config.name
    elif from_key:
        assert interactive, "Interactive mode must be enabled for --from-key"
        encoded_str = typer.prompt("Enter encoded string", hide_input=True)
        gen_config = Configuration.reinflate(encoded_str)
        name = gen_config.name

    else:
        if interactive:
            if name is None:
                name = typer.prompt(
                    "Enter configuration name", default=name)
                assert name is not None, "Configuration name must be specified"
                check_for_overwrite(name)
            db_host = typer.prompt(
                f"Enter {APP_NAME} host name", default=db_host)
            db_port = typer.prompt(
                f"Enter {APP_NAME} port number", default=db_port)
            db_username = typer.prompt(
                f"Enter {APP_NAME} username", default=db_username)
            db_password = typer.prompt(
                f"Enter {APP_NAME} password", hide_input=True, default=db_password)
            db_use_rest = typer.confirm(
                f"Use REST [Note: Only if ApertureDB is setup to receive HTTP requests]", default=db_use_rest)
            db_use_ssl = typer.confirm(
                f"Use SSL [Note: ApertureDB's defaults do not allow non SSL traffic]", default=db_use_ssl)

        gen_config = Configuration(
            name=name,
            host=db_host,
            port=db_port,
            username=db_username,
            password=db_password,
            use_ssl=db_use_ssl,
            use_rest=db_use_rest
        )

    assert name is not None, "Configuration name must be specified"
    configs[name] = gen_config
    if active:
        configs["active"] = name
    else:
        configs["active"] = ac

    with open(config_path.as_posix(), "w") as config_file:
        config_file.write(json.dumps(configs, indent=2, cls=ObjEncoder))


@app.command()
def activate(
        name: Annotated[str, typer.Argument(help="Name of this configuration for easy reference")],
        as_global: Annotated[bool, typer.Option(help="Project level vs global level")] = True):
    """
    Set the default configuration.
    """
    global_config_path = _config_file_path(True)
    gc, ga = get_configurations(global_config_path)

    config_path = _config_file_path(as_global)
    configs = {}
    try:
        configs, ac = get_configurations(config_path.as_posix())
        if name not in configs and name not in gc:
            console.log(f"Configuration {name} not found")
            raise typer.Exit(code=2)
        configs["active"] = name
    except FileNotFoundError:
        check_configured(as_global=False) or \
            check_configured(as_global=True, show_error=True)
    except json.JSONDecodeError:
        check_configured(as_global=False) or \
            check_configured(as_global=True, show_error=True)

    with open(config_path.as_posix(), "w") as config_file:
        config_file.write(json.dumps(configs, indent=2, cls=ObjEncoder))


@app.command()
def get_key(name: Annotated[str, typer.Argument(
        help="Name of the configuration to get a key for")] = None,
        user: Annotated[str, typer.Option(
            help="User to get a key for (default is config user)")] = None,
        as_global: Annotated[bool, typer.Option(help="Project level vs global level")] = True):
    """
    Makes a token from the configuration
    """

    config_path = _config_file_path(as_global)
    configs = {}
    user_key = None
    try:
        configs, active = get_configurations(config_path.as_posix())
        if not active and name is None:
            console.log(
                f"No configuration specified and no active configuration found")
            raise typer.Exit(code=2)
        if name is None:
            name = active
        if name not in configs and name not in gc:
            console.log(f"Configuration {name} not found")
            raise typer.Exit(code=2)
        configs["active"] = active

        if user is None:
            key_user = configs[name].username
        else:
            key_user = user

        if configs[name].has_user_keys():
            user_key = configs[name].get_user_key(key_user)

        if user_key is None:
            conn = __create_connector(configs[name])

            user_key = keys.generate_user_key(conn, key_user)
            configs[name].add_user_key(key_user, user_key)
            with open(config_path.as_posix(), "w") as config_file:
                config_file.write(json.dumps(
                    configs, indent=2, cls=ObjEncoder))
    except FileNotFoundError:
        check_configured(as_global=False) or \
            check_configured(as_global=True, show_error=True)
    except json.JSONDecodeError:
        check_configured(as_global=False) or \
            check_configured(as_global=True, show_error=True)

    print(f"{user_key}")



from enum import Enum
from typing import Annotated

import typer

from aperturedb.cli.console import console

app = typer.Typer()

import aperturedb.cli.keys as keys
import aperturedb.cli.tokens as tokens
app.add_typer(keys.app, name="keys",
              help="Manage Aperturedb keys")
app.add_typer(tokens.app, name="tokens",
              help="Manage database authentication tokens")


class CommandTypes(str, Enum):
    STATUS = "status"
    SUMMARY = "summary"
    REMOVE_ALL = "remove_all"
    REMOVE_INDEXES = "remove_indexes"


def confirm(command: CommandTypes, force: bool):
    if force:
        return True
    console.print("Danger", style="bold red")
    console.log(f"This will execute {command}.")
    response = typer.prompt("Are you sure you want to continue? [y/N]")
    if response.lower() != "y":
        typer.echo("Aborting...")
        raise typer.Abort()
    return True


@app.command(help="Execute a command on the database")
def execute(command: CommandTypes,
            force: Annotated[bool, typer.Option(help="Do not confirm")] = False):

    from aperturedb.Utils import Utils
    from aperturedb.CommonLibrary import create_connector

    utils = Utils(create_connector())
    available_commands = {
        CommandTypes.STATUS: lambda: print(utils.status()),
        CommandTypes.SUMMARY: utils.summary,
        CommandTypes.REMOVE_ALL: lambda: confirm(
            CommandTypes.REMOVE_ALL, force) and utils.remove_all_objects(),
        CommandTypes.REMOVE_INDEXES: lambda: confirm(
            CommandTypes.REMOVE_INDEXES, force) and utils.remove_all_indexes(),
    }

    available_commands[command]()


class LogLevel(str, Enum):
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"


@app.command()
def log(
    message: Annotated[str, typer.Argument(help="The message to log")],
    level: LogLevel = LogLevel.INFO
):
    """
    Log a message to the user log.

    This is useful because it can later be seen in Grafana, not only as log entries in the
    ApertureDB Logging dashboard, but also as event markers in the ApertureDB Status dashboard.
    """
    from aperturedb.Utils import Utils
    from aperturedb.CommonLibrary import create_connector

    utils = Utils(create_connector())
    utils.user_log_message(message, level=level.value)


@app.command()
def visualize_schema(
    filename: str = "schema",
    format: str = "png"
):
    """
    Visualize the schema of the database.

    This will create a file with the schema of the database in the specified format.

    Relies on graphviz to be installed.
    """
    from aperturedb.Utils import Utils
    from aperturedb.CommonLibrary import create_connector

    utils = Utils(create_connector())
    s = utils.visualize_schema()
    result = s.render(filename, format=format)
    print(result)



from enum import Enum
from typing import Annotated

import typer

from aperturedb.cli.console import console
from aperturedb.CommonLibrary import create_connector, execute_query
from aperturedb.Utils import Utils

app = typer.Typer()


@app.command(help="List User Authentication Tokens")
def list(user: Annotated[str, typer.Argument(help="The user the display tokens for")]):
    token_list_query = [{"GetUserDetails": {"username": user}}]
    client = create_connector()
    result, response, blobs = execute_query(
        client=client,
        query=token_list_query,
        blobs=[])
    utokens = response[0]['GetUserDetails']['tokens']
    if len(utokens) == 0:
        console.log(f"No Tokens for {user}")
    else:
        console.log(utokens)


@app.command(help="Generate an Authentication token for a user")
def generate():
    conn = create_connector()
    u = Utils(conn)
    token = u.generate_token()
    print(f"{token}")
    return token


@app.command(help="Assign an Authentication token to a user")
def assign(user: Annotated[str, typer.Argument(help="user to assign the token to")],
           token: Annotated[str, typer.Argument(help="Token to be assigned")]):
    conn = create_connector()
    u = Utils(conn)
    try:
        u.assign_token(user, token)
        console.log(f"Assigned token to {user}")
    except Exception as e:
        console.log(f"Failed to assign token: {e}", style="red")


@app.command(help="Remove an Authentication token from a user")
def remove(user: Annotated[str, typer.Argument(help="User to remove a token from")],
           token: Annotated[str, typer.Argument(help="Token to be removed")]):
    conn = create_connector()
    u = Utils(conn)
    try:
        u.remove_token(user, token)
        console.log("Action complete")
    except Exception as e:
        console.log(f"Failed to remove token: {e}", style="red")



import os
import stat
import errno
import fuse
from fuse import Fuse

from aperturedb.Images import Images
import logging
from threading import Lock
import json
from tqdm import tqdm

logger = logging.getLogger(__name__)

if not hasattr(fuse, '__version__'):
    raise RuntimeError(
        "your fuse-py doesn't know of fuse.__version__, probably it's too old.")

fuse.fuse_python_api = (0, 2)

meta_info = 'labels.json'


def generate_coco_meta_data(images: Images):
    """
    This function generates the meta data for the COCO dataset
    details can be found here: https://cocodataset.org/#format-data
    """
    properties = images.get_properties(images.get_props_names())

    # Hardcoded for now
    meta_licenses = [
        {
            "id": 1,
            "name": "Attribution-NonCommercial-ShareAlike License",
            "url": "http://creativecommons.org/licenses/by-nc-sa/2.0/",
        }
    ]

    meta_categories = [{
        "id": 1,
        "name": "unknown",
        "supercategory": "labels"
    }, {
        # Hardcoded for now
        "id": 2,
        "name": "Face points",
        "keypoints": ["lefteye", "righteyee", "nose", "leftmouth", "rightmouth"],
        "skeleton": [[0, 1, 3, 4], [2, 3, 4]]
    }
    ]

    meta_images = [{
        "id": id,
        "licensce": 1,
        "file_name": f"{id}.jpg",
        "height": properties[id]["adb_image_height"],
        "width": properties[id]["adb_image_width"],
    } for ind, id in enumerate(images.images_ids)]

    # Add attached bounding boxes
    categories = []
    meta_annotations = []

    for ind, image_id in tqdm(enumerate(images.images_ids)):
        if image_id in images.images_bboxes:
            for bidx, box in enumerate(images.images_bboxes[image_id]['bboxes']):
                annotation = {
                    "id": 2 * bidx,
                    "image_id": image_id,
                    "category_id": 1,
                    "bbox": [
                        int(box[0][0]),
                        int(box[0][1]),
                        int(box[1][0] - box[0][0]),
                        int(box[2][1] - box[0][1])
                    ]
                }
                meta_annotations.append(annotation)
                label = images.images_bboxes[image_id]['tags'][bidx]
                if label not in categories:
                    categories.append(label)

                # Figure the category id (offset by 1 for the unknown category)
                annotation["category_id"] = categories.index(label) + 3

        if image_id in images.images_polygons:
            for bidx, seg in enumerate(images.images_polygons[image_id]['polygons']):
                xs = [x for coords in seg[0]
                      for i, x in enumerate(coords) if i % 2 == 0]
                ys = [y for coords in seg[0]
                      for i, y in enumerate(coords) if i % 2 == 1]
                annotation = {
                    "id": 2 * bidx + 1,
                    "image_id": image_id,
                    "category_id": 1,
                    "segmentation": [[x for coords in seg[0] for x in coords]],
                    "tags": [k for k in properties[image_id] if properties[image_id][k] == 1],
                    "bbox": [
                        int(min(xs)),
                        int(min(ys)),
                        int(max(xs) - min(xs)),
                        int(max(ys) - min(ys))
                    ]
                }
                meta_annotations.append(annotation)
                label = images.images_polygons[image_id]['tags'][bidx]
                if label not in categories:
                    categories.append(label)

                # Figure the category id (offset by 1 for the unknown category)
                annotation["category_id"] = categories.index(label) + 3

        # Add attached keypoints. This is very DS specific, till we figure out
        # how to make this generic
        if "keypoints" in properties[image_id] and properties[image_id]["keypoints"] != None:
            ckps = []
            kps = properties[image_id]["keypoints"].split(" ")[1:]
            for i in range(0, len(kps), 2):
                ckps.append(float(kps[i]))
                ckps.append(float(kps[i + 1]))
                ckps.append(2)
            meta_annotations.append(
                {
                    "id": 2 * ind + 1,
                    "image_id": image_id,
                    "keypoints": ckps,
                    "num_keypoints": 5,
                    "category_id": 2
                }
            )

    for i, c in enumerate(categories):
        meta_categories.append({
            "id": i + 3,
            "name": c,
            "supercategory": "labels"
        })

    meta_data = bytes(json.dumps({
        "licenses": meta_licenses,
        "categories": meta_categories,
        "images": meta_images,
        "annotations": meta_annotations
    }, indent=2), 'utf-8')

    return meta_data, properties


class ADStat(fuse.Stat):
    def __init__(self):
        self.st_mode = 0
        self.st_ino = 0
        self.st_dev = 0
        self.st_nlink = 0
        self.st_uid = 0
        self.st_gid = 0
        self.st_size = 0
        self.st_atime = 0
        self.st_mtime = 0
        self.st_ctime = 0


class ADFS(Fuse):
    def __init__(self, images: Images, *args, **kw):
        super().__init__(*args, **kw)
        self._images = images
        self.iolock = Lock()
        self.stats = {}

        for i, img in enumerate(images.images_ids):
            logger.debug(img)
            try:
                images.get_bboxes_by_index(i)
                images.retrieve_polygons(i)
            except Exception as e:
                print(images.images_bboxes)
                raise

        self.meta_data, self.properties = generate_coco_meta_data(images)
        logger.info(f"Meta data generated with length = {len(self.meta_data)}")

    def getattr(self, path):
        logger.info(f"getattr: {path=}")
        st = ADStat()
        if path == '/':
            st.st_mode = stat.S_IFDIR | 0o755
            st.st_nlink = 2
        elif path == f"/data":
            st.st_mode = stat.S_IFDIR | 0o755
            st.st_nlink = 2
        elif path == f"/{meta_info}":
            st.st_mode = stat.S_IFREG | 0o444
            st.st_nlink = 1
            st.st_size = len(self.meta_data)
        else:
            st.st_mode = stat.S_IFREG | 0o444
            st.st_nlink = 1
            st.st_size = self.stats[path]
        logger.info(
            f"getattr: {path=}, {st.st_mode=}, {st.st_nlink=}, {st.st_size=}")
        return st

    def readdir(self, path, offset):
        if path == '/':
            direntries = [
                ('.', 0),
                ('..', 4096),
                ('data', 4096),
                (meta_info, len(self.meta_data))
            ]
        elif path == '/data':
            filegen = set()
            for iid in self.properties:
                filegen.add((f"{iid}.jpg",
                             self.properties[iid]["adb_image_size"]))

            direntries = [('.', 4096), ('..', 4096)] + [f for f in filegen]

        logger.info(f"readdir: {path=}, {len(direntries)=}")
        for r in direntries:
            self.stats[os.path.join(path, r[0])] = r[1]
            yield fuse.Direntry(r[0])

    def open(self, path, flags):
        logger.info(f"path = {path}, flags = {flags}")
        accmode = os.O_RDONLY | os.O_WRONLY | os.O_RDWR
        if (flags & accmode) != os.O_RDONLY:
            return -errno.EACCES

    def read(self, path, size, offset):
        logger.info(f"path = {path}, size = {size}, offset = {offset}")
        filename = os.path.basename(path)
        logger.info(f"Filename = {filename}")
        if filename == meta_info:
            try:
                img = self.meta_data
                slen = len(img)
            except Exception as e:
                logger.exception(e)
        else:
            try:
                image_id = filename[:-4]
                idx = self._images.images_ids.index(image_id)
                logger.info(f"Image id = {image_id}, index = {idx}")
                self.iolock.acquire()
                img = self._images.get_image_by_index(idx)
                self.iolock.release()
                slen = len(img)
                logger.debug(f"Type = {type(img)}, len = {slen}")
            except Exception as e:
                logger.exception(e)
                logger.error("Error occured")

        if offset < slen:
            if offset + size > slen:
                size = slen - offset
            buf = img[offset:offset + size]
        else:
            buf = b''
        return buf


def mount_images_from_aperturedb(images: Images):
    server = ADFS(
        version="%prog " + fuse.__version__,
        usage=fuse.Fuse.fusage,
        dash_s_do='setsingle',
        images=images)
    server.parse(errex=1)
    server.main()



import hashlib
from aperturedb.Subscriptable import Subscriptable
from aperturedb.transformers.transformer import Transformer
from .clip import generate_embedding, descriptor_set


class CLIPPyTorchEmbeddings(Transformer):
    """
    Generates the embeddings for the images using the CLIP Pytorch model.
    https://github.com/openai/CLIP
    """

    def __init__(self, data: Subscriptable, **kwargs) -> None:
        """
        Args:
            data: Subscriptable object
            search_set_name: Name of the [descriptorset](/query_language/Reference/descriptor_commands/desc_commands/AddDescriptor) to use for the search.
        """
        self.search_set_name = kwargs.pop(
            "search_set_name", descriptor_set)
        super().__init__(data, **kwargs)

        # Let's sample some data to figure out the descriptorset we need.
        if len(self._add_image_index) > 0:
            sample = generate_embedding(self.data[0][1][0])
            utils = self.get_utils()
            utils.add_descriptorset(
                self.search_set_name, dim=len(sample) // 4, metric=["CS"])

    def getitem(self, subscript):
        x = self.data[subscript]

        for ic in self._add_image_index:
            serialized = generate_embedding(x[1][ic])
            # If the image already has an image_sha256, we use it.
            image_sha256 = x[0][ic]["AddImage"].get("properties", {}).get(
                "adb_image_sha256", None)
            if not image_sha256:
                image_sha256 = hashlib.sha256(x[1][ic]).hexdigest()
            x[1].append(serialized)
            x[0].append(
                {
                    "AddDescriptor": {
                        "set": self.search_set_name,
                        "properties": {
                            "image_sha256": image_sha256,
                        },
                        "if_not_found": {
                            "image_sha256": ["==", image_sha256],
                        },
                        "connect": {
                            "ref": x[0][ic]["AddImage"]["_ref"]
                        }
                    }
                })
        return x



from dataclasses import dataclass

import json
import re
from base64 import b64encode, b64decode

APERTURE_CLOUD = ".cloud.aperturedata.io"
AD_CLOUD_SCHEME = "adbc://"


@dataclass(repr=False)
class Configuration:
    """
    **Configuration object for ApertureDB sdk to be able to connect to ApertureDB**
    """
    host: str
    port: int
    username: str
    password: str
    name: str
    use_ssl: bool = True
    use_rest: bool = False
    use_keepalive: bool = True
    retry_interval_seconds: int = 1
    # Max number of attempts to retry the initial connection (0 means infinite)
    # This is useful when the aperturedb server is not ready yet.
    retry_max_attempts: int = 3
    token: str = None
    user_keys: dict = None

    def __repr__(self) -> str:
        mode = "REST" if self.use_rest else "TCP"
        auth_mode = "token" if self.token is not None else "password"
        return f"[{self.host}:{self.port} as {self.username} using {mode} with SSL={self.use_ssl} auth={auth_mode}]"

    def deflate(self) -> list:
        deflate_version = 1
        host = self.host
        if host.endswith(APERTURE_CLOUD):
            host = "adbc://{}".format(host[:-1 * len(APERTURE_CLOUD)])
        as_list = [deflate_version, host, self.port, self.username, self.password, self.name, int(self.use_ssl),
                   int(self.use_rest), int(self.use_keepalive),
                   self.retry_interval_seconds, self.retry_max_attempts]
        simplified = json.dumps(as_list)
        encoded_key = b64encode(simplified.encode('utf-8')).decode('utf-8')
        return encoded_key

    def has_user_keys(self) -> bool:
        return self.user_keys is not None

    def add_user_key(self, user, key):
        if self.user_keys is None:
            self.user_keys = dict()
        if not user in self.user_keys:
            self.user_keys[user] = []
        self.user_keys[user].insert(0, key)

    def get_user_key(self, for_user: str) -> str:
        if self.user_keys is None or not (for_user in self.user_keys) \
                or len(self.user_keys[for_user]) == 0:
            return None
        return self.user_keys[for_user][0]

    def set_user_keys(self, keys: dict) -> None:
        self.user_keys = keys

    @classmethod
    def create_web_token(cls, host: str, port: int,  token_string: str) -> None:
        if token_string.startswith("adbp_"):
            token_string = token_string[5:]

        if host.endswith(APERTURE_CLOUD):
            host = host[:-1 * len(APERTURE_CLOUD)]
            m = re.match("(.*)\.farm(\d+)$", host)
            if m is not None:
                host = "{}.{}".format(m.group(1), int(m.group(2)))

            host = "a://{}".format(host)

        if port == 55555:
            web_token_json = [2, host, token_string]
        else:
            web_token_json = [2, host, port, token_string]
        simplified = json.dumps(web_token_json)
        encoded = b64encode(simplified.encode('utf-8')).decode('utf-8')
        return encoded

    @classmethod
    def reinflate(cls, encoded_str: list) -> object:
        try:
            decoded = b64decode(encoded_str.encode('utf-8'))
            as_list = json.loads(decoded.decode('utf-8'))
        except:
            raise Exception(
                "Unable to make configuration from the provided string")
        version = as_list[0]
        if version not in (1, 2):
            raise ValueError("version identifier of configuration was"
                             f"g{as_list[0]}, which is not supported")
        if version == 1:
            host, port, username, password, name, use_ssl, \
                use_rest, use_keepalive, retry_interval_seconds, \
                retry_max_attempts = as_list[1:]
            if host.startswith(AD_CLOUD_SCHEME):
                host = host[len(AD_CLOUD_SCHEME):] + APERTURE_CLOUD
            use_ssl = bool(use_ssl)
            use_rest = bool(use_rest)
            use_keepaliave = bool(use_keepalive)
            c = Configuration(
                host, port, username, password, name, use_ssl, use_rest, use_keepalive,
                retry_interval_seconds, retry_max_attempts)
        elif version == 2:
            host = as_list[1]
            if host.startswith("a://"):
                m = re.match("a://([^.]*)\.(\d+)", host)
                host = "{}.farm{:04d}.cloud.aperturedata.io".format(
                    m.group(1), int(m.group(2)))

            cur_arg = 2
            # second arg is port
            if isinstance(as_list[2], int):
                cur_arg = 3
            else:
                port = 55555

            name = "default"

            username = None
            password = None
            token = None
            # if 1 argument left, treat as token.
            if len(as_list) - cur_arg == 1:
                token = "adbp_" + as_list[cur_arg]
            else:
                username = as_list[cur_arg]
                password = as_list[cur_arg + 1]

            c = Configuration(host, port, username,
                              password, name, token = token)
        return c



from enum import Enum
from typing import Annotated

import typer

from aperturedb.cli.console import console
from aperturedb.Configuration import Configuration
from aperturedb.Connector import Connector
from aperturedb.CommonLibrary import create_connector
from aperturedb.Utils import Utils

app = typer.Typer()


@app.command(help="Create Key for a user")
def generate(user: Annotated[str, typer.Argument(help="The user to generate a key for")]):
    conn = create_connector()
    key = generate_user_key(conn, user)
    console.log(f"Key for {user} is", key, highlight=False)


def generate_user_key(conn: Connector, user: str):
    u = Utils(conn)
    token = u.generate_token()
    u.assign_token(user, token)
    key = Configuration.create_web_token(
        conn.config.host, conn.config.port, token)
    return key
